/*
* strict_chain.c - trim a blastz output file so that the alignments satisfy
* a strict chaining condition.
*
* The command syntax is either "strict_chain alignment_file" or just
* "strict_chain"; the second form reads stdin.  If the final argument is
* "verbose", it is interpreted as a request that information about how the
* alignments were modified is to be written to stderr.
*/

#include "lib.h"

#define IGNORE 20	// discard alignments shorter than this

static int verbose = 0;

enum {
  BUF_SIZE=500,
};

/* each gap-free segment of an alignment is stored */
typedef struct seg {
	int beg1, beg2, end1, end2, pct;
	struct seg *next, *prev;
} SEG;

/* each alignment is stored as a header and a list of segments */
typedef struct align {
	int beg1, beg2, end1, end2, score;
	int keep;
	SEG *first, *last;
} ALIGN;

/* The input is stored as a linked list of arbitrary lines, with interspersed
alignments */
typedef struct line_rec {
	char *line;
	ALIGN *align;
	struct line_rec *next;
} LINE_REC, *LINE_PTR;

static void discard(ALIGN *a, int loc) {
	if (verbose)
		fprintf(stderr, "%d: discard alignment %d-%d\n",
		  loc, a->beg1, a->end1);
	a->keep = 0;
}

static int covered(ALIGN *a, ALIGN *b) {
	int extend_left, extend_right;

	extend_left = MAX(b->beg1 - a->beg1, b->beg2 - a->beg2);
	extend_right = MAX(a->end1 - b->end1, a->end2 - b->end2);
	if (extend_left <= IGNORE && extend_right <= IGNORE) {
		discard(a, 0);
		return 1;
	}
	return 0;
}

// approximate 100 * number of matches in an interval in seq1 of an alignment
static int match1(ALIGN *a, int beg, int end) {
	int m100, len;
	SEG *s;

	if (beg > end)
		return 0;
	if (beg < a->beg1 || end > a->end1)
		fatal("match1: bad range");
	m100 = 0;
	for (s = a->first; s != NULL; s = s->next) {
		len = MIN(s->end1, end) - MAX(s->beg1, beg) + 1;
		if (len > 0)
			m100 += (len * s->pct);
	}
	return m100;
}

// approximate 100 * number of matches in an interval in seq2 of an alignment
static int match2(ALIGN *a, int beg, int end) {
	int m100, len;
	SEG *s;

	if (beg > end)
		return 0;
	if (beg < a->beg2 || end > a->end2)
		fatal("match2: bad range");
	m100 = 0;
	for (s = a->first; s != NULL; s = s->next) {
		len = MIN(s->end2, end) - MAX(s->beg2, beg) + 1;
		if (len > 0)
			m100 += (len * s->pct);
	}
	return m100;
}

void overlap(ALIGN *a, ALIGN *b)
{
	SEG *s;
	int x, y, z, beg_olap1, end_olap1, beg_olap2, end_olap2, am1, am2, bm1,
	  bm2, amatches, bmatches;

	if (verbose) {
		fprintf(stderr,
		   "alignment at %d-%d overlaps alignment at %d-%d",
			a->beg1, a->end1, b->beg1, b->end1);
		if (a->end1 < b->beg1)
			fprintf(stderr, " in 2nd sequence");
		fputc('\n', stderr);
	}
	if (a->beg1 > b->beg1) {
		ALIGN *tmp = a;
		a = b;
		b = tmp;
	}
	beg_olap1 = b->beg1;
	end_olap1 = MIN(a->end1, b->end1);
	beg_olap2 = MAX(a->beg2, b->beg2);
	end_olap2 = MIN(a->end2, b->end2);
	am1 = match1(a, beg_olap1, end_olap1);
	am2 = match2(a, beg_olap2, end_olap2);
	amatches = MAX(am1, am2);
	bm1 = match1(b, beg_olap1, end_olap1);
	bm2 = match2(b, beg_olap2, end_olap2);
	bmatches = MAX(bm1, bm2);
	if (amatches > bmatches) {
		// trim start of b to x+1 and y+1
		x = a->end1;
		y = a->end2;
		while (b->first &&
		    (b->first->end1 <= x || b->first->end2 <= y)) {
			/* discard b's first segment */
			if (b->first->next == NULL) {
				discard(b, 1);
				return;
			}
			b->beg1 = b->first->next->beg1;
			b->beg2 = b->first->next->beg2;
			s = b->first;
			b->first = b->first->next;
			free(s);
			if (b->first)
				b->first->prev = NULL;
		}
		if (b->first &&
		    (b->first->beg1 <= x || b->first->beg2 <= y)) {
			z = MAX(x+1-b->beg1, y+1-b->beg2);
			b->beg1 = (b->first->beg1 += z);
			b->beg2 = (b->first->beg2 += z);
		}
		if (verbose)
			fprintf(stderr,
			  "  trimmed start of second alignment to %d\n",
			  b->beg1);
	} else {	// trim end of a to x-1 and y-1
		x = b->beg1;
		y = b->beg2;
		while (a->last &&
		    (a->last->beg1 >= x || a->last->beg2 >= y)) {
			/* discard a's last segment */
			if (a->last->prev == NULL) {
				discard(a, 2);
				return;
			}
			a->end1 = a->last->prev->end1;
			a->end2 = a->last->prev->end2;
			s = a->last;
			a->last = a->last->prev;
			free(s);
			if (a->last)
				a->last->next = NULL;
		}
		if (a->last &&
		    (a->last->end1 >= x || a->last->end2 >= y)) {
			z = MAX(a->end1-x+1, a->end2-y+1);
			a->end1 = (a->last->end1 -= z);
			a->end2 = (a->last->end2 -= z);
		}
		if (verbose)
			fprintf(stderr,
			  "  trimmed end of first alignment to %d\n",
			  a->end1);
	}
}

void free_lines(LINE_PTR first_line)
{
	LINE_PTR p, q;
	SEG *s, *t;

	for (p = first_line; p; p = q) {
		q = p->next;
		if (p->line)
			free(p->line);
		else {
			for (s = p->align->first; s; s = t) {
				t = s->next;
				free(s);
			}
			free(p->align);
		}
		free(p);
	}
}

// we don't attempt a complete sort, but just look for the odd out-of-order
// alignment
LINE_PTR sort_align(LINE_PTR first_line) {
	LINE_PTR p, q;
	ALIGN *a;

	for (p = first_line; p != NULL; p = q) {
		while (p != NULL && p->line != NULL)
			p = p->next;
		if (p == NULL)
			break;
		q = p->next;
		while (q != NULL && (q->align == NULL || q->align->keep == 0))
			q = q->next;
		if (q == NULL)
			break;
		if (p->align->beg1 > q->align->beg1 ||
		    p->align->beg2 > q->align->beg2) {
			a = p->align;
			p->align = q->align;
			q->align = a;
		}
	}
	return first_line;
}

void print_lines(LINE_PTR p)
{
	ALIGN *a;
	SEG *s;

	for ( ; p; p = p->next)
		if (p->line)
			printf("%s", p->line);
		else if (p->align->keep) {
			a = p->align;
			printf("a {\n");
			printf("  s %d\n", a->score);
			printf("  b %d %d\n", a->beg1, a->beg2);
			printf("  e %d %d\n", a->end1, a->end2);
			for (s = a->first; s; s = s->next)
				printf("  l %d %d %d %d %d\n",
					s->beg1, s->beg2, s->end1, s->end2,
					s->pct);
			printf("}\n");
		}
}

int get_lines(FILE *ap, LINE_PTR *first_line_addr)
{
	char buf[BUF_SIZE];
	LINE_PTR last_line, new;
	ALIGN *newal;
	SEG *newseg;
	int nalign;

	if (fgets(buf, BUF_SIZE, ap) == NULL || strncmp(buf, "#:lav", 5))
		fatalf("arg1 is not a blastz file; first line = %s", buf);
	*first_line_addr = last_line = (LINE_PTR)ckalloc(sizeof(LINE_REC));
	last_line->align = NULL;
	last_line->line = copy_string(buf);
	nalign = 0;
	while (fgets(buf, BUF_SIZE, ap)) {
		new = (LINE_PTR)ckalloc(sizeof(LINE_REC));
		last_line = last_line->next = new;
		if (!same_string(buf, "a {\n")) {
			last_line->line = copy_string(buf);
			last_line->align = NULL;
		} else {
			++nalign;
			last_line->line = NULL;
			last_line->align = newal =
				(ALIGN *)ckalloc(sizeof(ALIGN));
			if  (fgets(buf, BUF_SIZE, ap) == NULL ||
			     sscanf(buf, "  s %d", &(newal->score)) != 1) 
				fatalf("no score: %s", buf);
			if  (fgets(buf, BUF_SIZE, ap) == NULL ||
			     sscanf(buf, "  b %d %d", &(newal->beg1),
						      &(newal->beg2)) != 2) 
				fatalf("no beginning: %s", buf);
			if  (fgets(buf, BUF_SIZE, ap) == NULL ||
			     sscanf(buf, "  e %d %d", &(newal->end1),
						      &(newal->end2)) != 2) 
				fatalf("no ending: %s", buf);
			newal->keep = 1;
			newal->first = newal->last = NULL;
			while (fgets(buf, BUF_SIZE, ap) && buf[0] != '}') {
				newseg = (SEG *)ckalloc(sizeof(SEG));
				if (sscanf(buf, "  l %d %d %d %d %d",
				    &(newseg->beg1), &(newseg->beg2),
				    &(newseg->end1), &(newseg->end2),
				    &(newseg->pct)) != 5)
					fatalf("Cannot get segment: %s", buf);
				if (newal->first == NULL) {
					newal->first = newseg;
					newseg->prev = newseg->next = NULL;
				} else {
					newseg->prev = newal->last;
					newal->last->next = newseg;
				}
				newseg->next = NULL;
				newal->last = newseg;
			}
		}
	}
	last_line->next = NULL;

	return nalign;
}


int main(int argc, char **argv)
{
	FILE *fp;
	LINE_PTR first_line, p;
	int i, j, n, n_align, *min_later_beg1, *min_later_beg2;
	ALIGN **al;

	argv0 = "strict_chain";
	if (same_string(argv[argc-1], "verbose")) {
		verbose = 1;
		--argc;
	}
	if (argc != 1 && argc != 2)
		fatalf("args = [blastz-output] [verbose]");

	fp = (argc == 2 ? ckopen(argv[1], "r") : stdin);

	n_align = get_lines(fp, &first_line);
	if (n_align == 0)
		fatal("no alignments");

	al = ckalloc(n_align*sizeof(ALIGN *));
	for (n = 0, p = first_line; p; p = p->next)
		if (p->align) {
			if (n >= n_align)
				fatal("Too many alignments.");
			al[n++] = p->align;
		}
	for (i = 1; i < n_align; ++i)
		if (al[i-1]->beg1 > al[i]->beg1)
			fatal("alignments out of order");
	min_later_beg1 = ckalloc(n_align*sizeof(int));
	min_later_beg2 = ckalloc(n_align*sizeof(int));
	min_later_beg1[n_align-1] = al[n_align-1]->beg1;
	min_later_beg2[n_align-1] = al[n_align-1]->beg2;
	for (i = n_align-2; i >= 0; --i) {
		min_later_beg1[i] = MIN(al[i]->beg1,min_later_beg1[i+1]);
		min_later_beg2[i] = MIN(al[i]->beg2,min_later_beg2[i+1]);
	}

	for (i = 0; i < n_align; ++i)
		for (j = i+1; j < n_align; ++j) {
			if (al[i]->keep == 0)
				break;
			if (al[j]->keep == 0)
				continue;
			if (min_later_beg1[j] > al[i]->end1 &&
			    min_later_beg2[j] > al[i]->end2)
				break;
			if (!covered(al[i], al[j]) && !covered(al[j], al[i]))
				overlap(al[i], al[j]);
		}
	for (i = 0; i < n_align; ++i)
		if (al[i]->end1 - al[i]->beg1 + 1 <= IGNORE)
			al[i]->keep = 0;

	first_line = sort_align(first_line);
 
	print_lines(first_line);
	if (fp != stdin)
		fclose(fp);
/*
	free_lines(first_line);
	free(al);
	free(min_later_beg1);
	free(min_later_beg2);
*/
	return 0;
}
